#!/usr/bin/env python
#
# test_property_string.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import fsleyes_props as props


def test_assign():
    class MyObj(props.HasProperties):
        val = props.String()

    obj = MyObj()


    # TODO


def test_length():
    class MyObj(props.HasProperties):
        val = props.String()



#!/usr/bin/env python
#
# __init__.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

from __future__ import print_function

import gc
import time

import wx

from  fsl.utils.platform import platform as fslplatform


def run_with_wx(func, *args, **kwargs):

    gc.collect()

    propagateRaise = kwargs.pop('propagateRaise', True)
    startingDelay  = kwargs.pop('startingDelay',  500)
    finishingDelay = kwargs.pop('finishingDelay', 500)
    callAfterApp   = kwargs.pop('callAfterApp',   None)

    result = [None]
    raised = [None]

    app    = [wx.App()]
    frame  = wx.Frame(None)
    panel  = wx.Panel(frame)
    sizer  = wx.BoxSizer(wx.VERTICAL)
    sizer.Add(panel, flag=wx.EXPAND, proportion=1)
    frame.SetSizer(sizer)
    frame.Layout()

    if callAfterApp is not None:
        callAfterApp()

    def wrap():

        try:
            if func is not None:
                result[0] = func(panel, *args, **kwargs)

        except Exception as e:
            print(e)
            raised[0] = e

        finally:
            def finish():
                frame.Destroy()
                app[0].ExitMainLoop()

            wx.CallLater(finishingDelay, finish)

    frame.Show()

    wx.CallLater(startingDelay, wrap)

    app[0].MainLoop()

    if raised[0] and propagateRaise:
        raise raised[0]

    del app[0]
    return result[0]


def addall(parent, widgets):
    sizer = wx.BoxSizer(wx.VERTICAL)
    parent.SetSizer(sizer)
    for w in widgets:
        sizer.Add(w, flag=wx.EXPAND, proportion=1)
    parent.Layout()
    parent.Refresh()
    realYield()



# Under GTK, a single call to
# yield just doesn't cut it
def realYield(centis=10):
    for i in range(int(centis)):
        wx.YieldIfNeeded()
        time.sleep(0.01)


# stype:
#   0 for single click
#   1 for double click
#   2 for separatemouse down/up events
def simclick(sim, target, btn=wx.MOUSE_BTN_LEFT, pos=None, stype=0):

    w, h = target.GetClientSize().Get()
    x, y = target.GetScreenPosition()

    if pos is None:
        pos = [0.5, 0.5]

    x += w * pos[0]
    y += h * pos[1]

    sim.MouseMove(x, y)
    wx.Yield()
    if   stype == 0: sim.MouseClick(btn)
    elif stype == 1: sim.MouseDblClick(btn)
    else:
        sim.MouseDown(btn)
        sim.MouseUp(btn)
    realYield()



def simtext(sim, target, text, enter=True):
    target.SetFocus()
    target.SetValue(text)

    # KeyDown doesn't seem to work
    # under docker/GTK so we have
    # to hack
    if enter and fslplatform.wxPlatform == fslplatform.WX_GTK:
        parent = target.GetParent()
        if type(parent).__name__ == 'FloatSpinCtrl':
            parent._FloatSpinCtrl__onText(None)
        else:
            sim.KeyDown(wx.WXK_RETURN)
    elif enter:
        sim.KeyDown(wx.WXK_RETURN)
    realYield()


def simkey(sim, target, key, down=True, up=False):
    if target is not None: target.SetFocus()
    if down:               sim.KeyDown(key)
    if up:                 sim.KeyUp(key)
    realYield()



#!/usr/bin/env python
#
# callqueue.py - A queue for calling functions sequentially.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`CallQueue` class, which is used by
:class:`.PropertyValue` instances to enqueue and execute property listener
callback functions.
"""


import logging

import six.moves.queue as queue

import fsl.utils.idle  as idle


log = logging.getLogger(__name__)


class Call(object):
    """A little class which is used to represent function calls that are
    on the queue.
    """

    def __init__(self, func, name, args, kwargs):
        self.func    = func
        self.name    = name
        self.args    = args
        self.kwargs  = kwargs
        self.execute = True

        # The CallQueue.dequeue method sets the
        # above execute attribute to False for
        # calls which are to be dequeued - this
        # causees the CallQueue.__call method
        # to skip over the call.


class CallQueue(object):
    """A queue of functions to be called. Functions can be enqueued via
    the :meth:`call` or :meth:`callAll` methods.
    """

    def __init__(self, skipDuplicates=False):
        """Create a ``CallQueue`` instance.

        If ``skipDuplicates`` is ``True``, a function which is already on
        the queue will be silently dropped if an attempt is made to add it
        again.

          .. note::

             The ``skipDuplicates`` test is based solely on the name of the
             function. This means that the ``CallQueue`` does not support
             enqueueing the same function with different arguments.

             Testing for function *and* argument equality is a difficult task:

               - I can't take the hash of the arguments, as I can't assume
                 that they are hashable (e.g.  ``numpy`` arrays).

               - I can't test for identity, as things which have the same
                 value may not have the same id (e.g. strings).

               - I can't test argument equality, because in some cases the
                 argument may be a mutable type (e.g. a ``list``), and its
                 value may have changed between the time the function was
                 queued, and the time it is called. *And* the arguments might
                 be big (again, ``numpy`` arrays), so an equality test could
                 be expensive.

             So this is quite a pickle. Something to come back to if things
             are breaking because of it.


        **Holding the queue**


        The :meth:`hold` method temporarily stops the ``CallQueue`` from
        queueing and executing functions. Any functions which are enqueued
        while the queue is held are kept in a separate queue. The queue is
        released via the :meth:`release` method, after which any held
        functions may be accessed via the :meth:`clearHeld` method (which
        also clears the internal queue of held functions). Once the queue
        has been released, these held functions can be re-queued as normal
        via the :meth:`call` or :meth:`callAll` methods.
        """

        # The queue is a queue of Call instances
        #
        # The queued dict contains mappings of
        # {name : [List of Call instances]}
        #
        self.__queue          = queue.Queue()
        self.__queued         = {}
        self.__skipDuplicates = skipDuplicates
        self.__calling        = False


        # Every call to hold will increment
        # this count, and every call to
        # release will decrement it.
        self.__holding        = 0


        # If the queue is being held, enqueued
        # functions are added to this list
        self.__held           = []


    @idle.mutex
    def dequeue(self, name):
        """If the specified function is on the queue, it is (effectively)
        dequeued, and not executed.

        If ``skipDuplicates`` is ``False``, and more than one function of
        the same name is enqueued, they are all dequeued.
        """

        # Get all calls with the specified name
        calls = self.__queued.get(name, [])

        # Set each of their execute flags to
        # False - the __call method will skip
        # over Call instances with execute=False
        for call in calls:

            self.__debug(call, 'Dequeueing function', 'from queue')
            call.execute = False

        # Check the held queue as well
        for call in self.__held:
            if call.name == name:
                self.__debug(call, 'Dequeueing held function', 'from queue')
                call.execute = False


    def call(self, func, name, *args, **kwargs):
        """Enqueues the given function, and calls all functions in the queue

        (unless the call to this method was as a result another function
        being called from the queue).
        """

        if self.__push(Call(func, name, args, kwargs)):
            self.__call()


    def callAll(self, funcs):
        """Enqueues all of the given functions, and calls all functions in
        the queue.

        (unless the call to this method was as a result another function
        being called from the queue).

        Assumes that the given ``funcs`` parameter is a list of
        ``(function, name, args, kwargs)`` tuples.
        """

        anyEnqueued = False

        for func in funcs:
            if self.__push(Call(*func)):
                anyEnqueued = True

        if anyEnqueued:
            self.__call()


    @idle.mutex
    def hold(self):
        """Holds the queue. For every call to ``hold``, the :meth:`release`
        method must be called once before the queue will be truly released.
        """
        self.__holding += 1


    @idle.mutex
    def release(self):
        """Releases the queue. """
        self.__holding -= 1


    @idle.mutex
    def clearHeld(self):
        """Clears and returns the list of held functions. """

        if self.__holding > 0:
            return []

        held = self.__held
        self.__held = []
        return [(c.func, c.name, c.args, c.kwargs) for c in held if c.execute]


    def __call(self):
        """Call all of the functions which are currently enqueued.

        This method is not re-entrant - if a call to one of the
        functions in the queue triggers another call to this method,
        this second call will return immediately without doing anything.
        """

        if self.__calling: return
        self.__calling = True

        while True:

            try:
                call = self.__pop()

                if not call.execute:
                    self.__debug(call, 'Skipping dequeued function')
                    continue

                self.__debug(call, 'Calling function')

                try:
                    call.func(*call.args, **call.kwargs)

                except Exception as e:
                    import traceback
                    log.warning('Function {} raised exception: {}'.format(
                        call.name, e), exc_info=True)
                    traceback.print_stack()

            except queue.Empty:
                break

        self.__calling = False


    @idle.mutex
    def __push(self, call):
        """Enqueues the given ``Call`` instance.

        If the queue has been held (see :meth:`hold`), the call is
        stored, and ``False`` is returned.

        If ``True`` was passed in for the ``skipDuplicates`` parameter
        during initialisation, and the function is already enqueued,
        it is not added to the queue, and this method returns ``False``.

        Otherwise, this method returnes ``True``.
        """

        enqueued = self.__queued.get(call.name, [])

        if self.__holding > 0:
            self.__debug(call, 'Holding function')
            self.__held.append(call)
            return False

        # Skip this function if there are already
        # functions in the queue with the same name
        if self.__skipDuplicates and (len(enqueued) > 0):
            self.__debug(call, 'Skipping function')
            return False

        self.__debug(call, 'Queueing function', 'to queue')

        self.__queue.put_nowait(call)
        self.__queued[call.name] = enqueued + [call]

        return True


    @idle.mutex
    def __pop(self):
        """Pops the next function from the queue and returns the ``Call``
        instance which encapsulates it.
        """

        call     = self.__queue.get_nowait()
        enqueued = self.__queued[call.name]

        # TODO shouldn't the call always
        # be at the end of the list?
        #
        # This will be a little bit faster
        # if you remove the call by index.
        enqueued.remove(call)

        if len(enqueued) == 0:
            self.__queued.pop(call.name)

        return call


    def __debug(self, call, prefix, postfix=None):
        """Prints a standardised log message."""

        if log.getEffectiveLevel() != logging.DEBUG:
            return

        funcName, modName = self.__getCallbackDetails(call.func)

        if postfix is None: postfix = ' '
        else:               postfix = ' {} '.format(postfix)

        log.debug('{} {} [{}.{}]{}({} in queue)'.format(
            prefix,
            call.name,
            modName,
            funcName,
            postfix,
            self.__queue.qsize()))


    def __getCallbackDetails(self, cb):
        """Returns the function name and module name of the given
        function reference. Used purely for debug log statements.
        """

        import inspect

        if log.getEffectiveLevel() != logging.DEBUG:
            return '', ''

        members = inspect.getmembers(cb)

        funcName  = ''
        modName   = ''

        for name, val in members:
            if name == '__func__':
                funcName  = val.__name__
                modName   = val.__module__
                break

        return funcName, modName



#!/usr/bin/env python
#
# widgets_bounds.py - Create widgets for modifying Bounds properties.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

"""This module provides the :func:`_Bounds` function, which is imported into
the :mod:`widgets` module namespace. It is separated purely to keep the
``widgets`` module file size down.
"""


import wx

import numpy as np

import fsleyes_widgets.rangeslider as rangeslider


def _Bounds(parent,
            hasProps,
            propObj,
            propVal,
            slider=True,
            spin=True,
            showLimits=True,
            editLimits=True,
            mousewheel=False,
            labels=None,
            spinWidth=None,
            **kwargs):
    """Creates and returns a panel containing sliders/spinboxes which
    allow the user to edit the low/high values along each dimension of the
    given :class:`.Bounds` property value.


    If both the ``slider`` and ``spin`` arguments are ``True``, a
    :class:`.RangeSliderSpinPanel` widget is returned; otherwise
    a :class:`.RangePanel` is returned.


    If both ``slider`` and ``spin`` are ``False``, a :exc:`ValueError` is
    raised.


    :arg slider:     Display slider widgets allowing the user to control the
                     bound values.

    :arg spin:       Display spin control widgets allowing the user to control
                     the bound values.

    :arg showLimits: Show the bound limits.

    :arg editLimits: Add buttons allowing the user to edit the bound limits.

    :arg mousewheel: If ``True``, mouse wheel events over the slider/spin
                     controls will change the bounds values.

    :arg labels:     A list of strings of length ``2 * ndims``, where ``ndims``
                     is the number of dimensions on the ``Bounds`` property;
                     the strings are used as labels on the widget.

    :arg spinWidth:  Desired spin control width. See the
                     :class:`.FloatSpinCtrl` class.

    See the :func:`.widgets._String` documentation for details on the other
    parameters.
    """

    ndims   = propObj._ndims
    real    = propObj._real
    clamped = propObj.getListType().getAttribute(None, 'clamped')

    panel = wx.Panel(parent)
    sizer = wx.BoxSizer(wx.VERTICAL)
    panel.SetSizer(sizer)

    if labels is None:
        labels = [None] * 2 * ndims

    for i in range(ndims):
        minDistance = propObj.getAttribute(hasProps, 'minDistance')
        minval      = propVal.getMin(i)
        maxval      = propVal.getMax(i)
        loval       = propVal.getLo(i)
        hival       = propVal.getHi(i)

        if minDistance is None: minDistance = 0
        if minval      is None: minval      = loval
        if maxval      is None: maxval      = hival

        if slider and spin:

            style = 0

            if mousewheel:  style |= rangeslider.RSSP_MOUSEWHEEL
            if not real:    style |= rangeslider.RSSP_INTEGER
            if showLimits:  style |= rangeslider.RSSP_SHOW_LIMITS
            if editLimits:  style |= rangeslider.RSSP_EDIT_LIMITS
            if not clamped: style |= rangeslider.RSSP_NO_LIMIT

            slider = rangeslider.RangeSliderSpinPanel(
                panel,
                minValue=minval,
                maxValue=maxval,
                lowValue=loval,
                highValue=hival,
                lowLabel=labels[i * 2],
                highLabel=labels[i * 2 + 1],
                minDistance=minDistance,
                spinWidth=spinWidth,
                style=style)

        elif slider or spin:

            style = 0

            if mousewheel:  style |= rangeslider.RP_MOUSEWHEEL
            if slider:      style |= rangeslider.RP_SLIDER
            if not real:    style |= rangeslider.RP_INTEGER
            if not clamped: style |= rangeslider.RP_NO_LIMIT

            slider = rangeslider.RangePanel(
                panel,
                minValue=minval,
                maxValue=maxval,
                lowValue=loval,
                highValue=hival,
                lowLabel=labels[i * 2],
                highLabel=labels[i * 2 + 1],
                minDistance=minDistance,
                spinWidth=spinWidth,
                style=style)

        else:
            raise ValueError('One of slider or spin must be True')

        _boundBind(hasProps, propObj, slider, propVal, i, editLimits)

        sizer.Add(slider, flag=wx.EXPAND)

    return panel


def _boundBind(hasProps, propObj, sliderPanel, propVal, axis, editLimits):
    """Called by the :func:`_Bounds` function.

    Binds the given :class:`.RangeSliderSpinPanel` or :class:`.RangePanel` to
    one axis of the given :class:`.BoundsValueList` so that changes in one are
    propagated to the other.

    :arg sliderPanel: The :class:`.RangeSliderSpinPanel`/:class:`.RangePanel`
                      instance.

    :arg axis:        The 0-indexed axis of the :class:`.Bounds` value.

    See :func:`_Bounds` for details on the other arguments.
    """


    lowProp    = propVal.getPropertyValueList()[axis * 2]
    highProp   = propVal.getPropertyValueList()[axis * 2 + 1]

    lowName     = 'BoundBind_{}_{}'.format(id(sliderPanel), id(lowProp))
    highName    = 'BoundBind_{}_{}'.format(id(sliderPanel), id(highProp))
    lowAttName  = 'BoundBindAtt_{}_{}'.format(id(sliderPanel), id(lowProp))
    highAttName = 'BoundBindAtt_{}_{}'.format(id(sliderPanel), id(highProp))

    # Called when the low or high PV changes
    def guiUpdate(*a):
        lowVal  = lowProp.get()
        highVal = highProp.get()

        if np.all(np.isclose(sliderPanel.GetRange(), (lowVal, highVal))):
            return

        sliderPanel.SetRange(lowVal, highVal)

    # Called on a rangeslider.EVT_LOW_RANGE event
    def lowPropUpdate(ev):
        lowProp.disableListener(lowName)
        lowProp.set(ev.low)
        lowProp.enableListener(lowName)
        ev.Skip()

    # Called on a rangeslider.EVT_HIGH_RANGE event
    def highPropUpdate(ev):
        highProp.disableListener(highName)
        highProp.set(ev.high)
        highProp.enableListener(highName)
        ev.Skip()

    # Called on a rangeslider.EVT_RANGE event
    def propUpdate(ev):
        lowProp .disableListener(lowName)
        highProp.disableListener(highName)
        propVal.setRange(axis, ev.low, ev.high)
        lowProp .enableListener(lowName)
        highProp.enableListener(highName)

    # Called when any attributes of the
    # individual bounds property values change
    def updateSliderRange(ctx, att, *a):

        if att not in ('minval', 'maxval'):
            return

        minval = propVal.getMin(axis)
        maxval = propVal.getMax(axis)

        sliderPanel.SetLimits(minval, maxval)

    # Called on rangeslider.EVT_RANGE_LIMIT events
    def updatePropRange(ev):
        lowProp .disableAttributeListener(lowName)
        highProp.disableAttributeListener(highName)
        propVal.setLimits(axis, ev.min, ev.max)
        lowProp .enableAttributeListener(lowName)
        highProp.enableAttributeListener(highName)
        ev.Skip()

    sliderPanel.Bind(rangeslider.EVT_RANGE,      propUpdate)
    sliderPanel.Bind(rangeslider.EVT_LOW_RANGE,  lowPropUpdate)
    sliderPanel.Bind(rangeslider.EVT_HIGH_RANGE, highPropUpdate)

    lowProp .addListener(lowName,  guiUpdate, weak=False)
    highProp.addListener(highName, guiUpdate, weak=False)

    lowProp .addAttributeListener(lowAttName,  updateSliderRange, weak=False)
    highProp.addAttributeListener(highAttName, updateSliderRange, weak=False)

    if editLimits:
        sliderPanel.Bind(rangeslider.EVT_RANGE_LIMIT, updatePropRange)

    def onDestroy(ev):
        ev.Skip()
        if ev.GetEventObject() is not sliderPanel:
            return

        lowProp .removeListener(         lowName)
        highProp.removeListener(         highName)
        lowProp .removeAttributeListener(lowAttName)
        highProp.removeAttributeListener(highAttName)

    sliderPanel.Bind(wx.EVT_WINDOW_DESTROY, onDestroy)



#!/usr/bin/env python
#
# test_property_types.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import numpy as np

import fsleyes_props as props


def test_define_and_assign():
    class Thing(props.HasProperties):

        myobject     = props.Object()
        mybool       = props.Boolean()
        myint        = props.Int()
        myreal       = props.Real()
        mypercentage = props.Percentage()
        mystring     = props.String()
        mychoice     = props.Choice(('1', '2', '3', '4', '5'))
        myfilepath   = props.FilePath()
        mylist       = props.List()
        mycolour     = props.Colour()
        mycolourmap  = props.ColourMap()
        mybounds     = props.Bounds(ndims=2)
        mypoint      = props.Point(ndims=2)
        myarray      = props.Array()

    t = Thing()

    t.myobject     = '12345'
    t.mybool       = False
    t.myint        = 12345
    t.myreal       = 0.12345
    t.mypercentage = 12.345
    t.mystring     = '12345'
    t.mychoice     = '5'
    t.myfilepath   = '/path/to/nonexistent/file.txt'
    t.mylist       = [1, 2, 3, 4, 5]
    t.mycolour     = (0.2, 0.3, 0.5, 1.0)
    t.mycolourmap  = 'cool'
    t.mybounds     = [0, 1, 0, 1]
    t.mypoint      = [0, 1]
    t.myarray      = np.eye(4)



#!/usr/bin/env python
#
# widgets_list.py - A widget for editing a List property.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""A widget for editing a :class:`.List` property.

.. warning:: The code in this module is old, and probably doesn't work. It
             will be  updated when I, or somebody else,  needs a list widget.
"""

import wx

import fsleyes_widgets.elistbox as elistbox
from . import properties_types  as ptypes
from . import                      widgets


def _pasteDataDialog(parent, hasProps, propObj):
    """Displays a dialog containing an editable text field, allowing the
    user to type/paste bulk data which will be used to populate the list
    (one line per item).

    :param parent:   parent GUI object

    :param hasProps: The :class:`.HasProperties` object which owns the
                     ``propObj``.

    :param propObj:  The :class:`.List` property object.
    """

    listObj  = getattr(hasProps, propObj.getLabel(hasProps))
    initText = '\n'.join([str(l).strip() for l in listObj])

    frame = wx.Dialog(parent,
                      style=wx.DEFAULT_DIALOG_STYLE |
                            wx.RESIZE_BORDER)
    panel = wx.Panel(frame)
    text  = wx.TextCtrl(panel, value=initText,
                        style=wx.TE_MULTILINE |
                              wx.TE_DONTWRAP  |
                              wx.VSCROLL      |
                              wx.HSCROLL)

    # ok/cancel buttons
    okButton     = wx.Button(panel, label="Ok")
    cancelButton = wx.Button(panel, label="Cancel")

    sizer = wx.GridBagSizer()

    panel.SetSizer(sizer)

    sizer.Add(text,         pos=(0, 0), span=(1, 2), flag=wx.EXPAND)
    sizer.Add(okButton,     pos=(1, 0), span=(1, 1), flag=wx.EXPAND)
    sizer.Add(cancelButton, pos=(1, 1), span=(1, 1), flag=wx.EXPAND)

    sizer.AddGrowableRow(0)
    sizer.AddGrowableCol(0)
    sizer.AddGrowableCol(1)

    def pasteIntoList():
        """
        Called when the user pushes 'Ok'. Copies whatever the
        user typed/pasted into the text area into the list, and
        closes the dialog window.
        """

        listData = text.GetValue().strip()
        listData = listData.split('\n')
        listData = [s.strip() for s in listData]

        setattr(hasProps, propObj.getLabel(hasProps), listData)
        frame.Close()

    okButton    .Bind(wx.EVT_BUTTON, lambda e: pasteIntoList())
    cancelButton.Bind(wx.EVT_BUTTON, lambda e: frame.Close())

    panel.Fit()
    panel.Layout()

    frame.ShowModal()


def _editListDialog(parent, hasProps, propObj):
    """A dialog which displays a widget for every item in the list, and
    which allows the user to adjust the number of items in the list.

    See the :func:`_pasteDataDialog` for details on the parameters.
    """

    # listObj is a properties_values.PropertyValueList object
    listObj  = getattr(hasProps, propObj.getLabel(hasProps))
    listType = propObj._listType
    propName = propObj.getLabel(hasProps)

    # min/max values for a spin box which allows the
    # user to select the number of items in the list
    minval = propObj.getAttribute(hasProps, 'minlen')
    maxval = propObj.getAttribute(hasProps, 'maxlen')

    if minval is None: minval = 0
    if maxval is None: maxval = 2 ** 31 - 1

    frame        = wx.Dialog(parent,
                             style=wx.DEFAULT_DIALOG_STYLE |
                                   wx.RESIZE_BORDER)
    numRowsPanel = wx.Panel(frame)
    entryPanel   = wx.ScrolledWindow(frame)
    okButton     = wx.Button(frame, label='Ok')

    # Spin box (and label) to adjust the
    # number of entries in the list
    numRowsSizer = wx.BoxSizer(wx.HORIZONTAL)
    numRowsPanel.SetSizer(numRowsSizer)

    numRowsLabel = wx.StaticText(numRowsPanel, label='Number of entries')
    numRowsCtrl  = wx.SpinCtrl(numRowsPanel,
                               min=minval,
                               max=maxval,
                               initial=len(listObj))

    numRowsSizer.Add(numRowsLabel, flag=wx.EXPAND, proportion=1)
    numRowsSizer.Add(numRowsCtrl,  flag=wx.EXPAND, proportion=1)

    listWidgets = []

    # Make a widget for every element in the list
    propVals = listObj.getPropertyValueList()
    for propVal in propVals:
        widget  = widgets.makeWidget(entryPanel, hasProps, propName)
        listWidgets.append(widget)

    frameSizer = wx.BoxSizer(wx.VERTICAL)
    frame.SetSizer(frameSizer)
    frameSizer.Add(numRowsPanel, flag=wx.EXPAND)
    frameSizer.Add(entryPanel,   flag=wx.EXPAND, proportion=1)
    frameSizer.Add(okButton,     flag=wx.EXPAND)

    entryPanelSizer = wx.BoxSizer(wx.VERTICAL)
    entryPanel.SetSizer(entryPanelSizer)
    for lw in listWidgets:
        entryPanelSizer.Add(lw, flag=wx.EXPAND)


    def changeNumRows(*args):
        """
        Called when the numRows variable is changed (via the numRows
        spinbox). Adds or removes the last  item from the property
        list, and corresponding widget from the window.
        """

        oldLen   = len(listObj)
        newLen   = numRowsCtrl.GetValue()

        # add rows
        default = listType.getAttribute(hasProps, 'default')
        while oldLen < newLen:

            # add a new element to the list
            listObj.append(default)

            # add a widget
            widg = widgets.makeWidget(entryPanel, hasProps, propName)
            listWidgets.append(widg)
            entryPanelSizer.Add(widg, flag=wx.EXPAND)

            oldLen = oldLen + 1

        # remove rows
        while oldLen > newLen:

            # kill the list item
            listObj.pop()

            # kill the widget
            widg = listWidgets.pop()

            entryPanelSizer.Remove(widg)

            widg.Destroy()

            oldLen = oldLen - 1
        entryPanel.Layout()
        entryPanel.Refresh()


    numRowsCtrl.Bind(wx.EVT_SPINCTRL, changeNumRows)
    okButton.Bind(wx.EVT_BUTTON, lambda e: frame.Close())

    frame.ShowModal()


def _listDialogWidget(parent, hasProps, propObj, propVal):
    """Creates and returns a GUI panel containing two buttons which, when
    pushed, display dialogs allowing the user to edit the values contained
    in the list.

    See the :func:`_pasteDataDialog` and :func:`_editListDialog` functions.
    """

    panel = wx.Panel(parent)

    # When the user pushes this button on the parent window,
    # a new window is displayed, allowing the user to edit
    # the values in the list individually
    editButton = wx.Button(panel, label='Edit')
    editButton.Bind(
        wx.EVT_BUTTON, lambda ev: _editListDialog(parent, hasProps, propObj))

    # When the user pushes this button, a new window is
    # displayed, allowing the user to type/paste bulk data
    pasteButton = wx.Button(panel, label='Paste data')
    pasteButton.Bind(
        wx.EVT_BUTTON, lambda ev: _pasteDataDialog(parent, hasProps, propObj))

    sizer = wx.BoxSizer(wx.HORIZONTAL)
    panel.SetSizer(sizer)

    sizer.Add(editButton,  flag=wx.EXPAND)
    sizer.Add(pasteButton, flag=wx.EXPAND)

    panel.Layout()

    return panel


def _listEmbedWidget(parent, hasProps, propObj, propVal):

    if not isinstance(propObj._listType, (ptypes.String,
                                          ptypes.Int,
                                          ptypes.Real)):
        raise RuntimeError('Unsupported property '
                           'type: {}'.format(propObj.__class__))

    widget = elistbox.EditableListBox(
        parent,
        ['{}'.format(v) for v in propVal],
        style=(elistbox.ELB_NO_ADD    |
               elistbox.ELB_NO_REMOVE |
               elistbox.ELB_EDITABLE))

    changeTriggeredByWidget = [False]

    def _listBoxMove(ev):
        changeTriggeredByWidget[0] = True
        propVal.move(ev.oldIdx, ev.newIdx)
        changeTriggeredByWidget[0] = False

    def _listBoxEdit(ev):
        changeTriggeredByWidget[0] = True
        propVal[ev.idx] = ev.label
        changeTriggeredByWidget[0] = False

    def _listChanged(*a):

        if changeTriggeredByWidget[0]:
            return

        sel = widget.GetSelection()
        widget.Clear()
        for val in propVal:
            widget.Append('{}'.format(val))

        if sel < widget.GetCount():
            widget.SetSelection(sel)
        else:
            widget.SetSelection(widget.GetCount() - 1)

    widget.Bind(elistbox.EVT_ELB_MOVE_EVENT,   _listBoxMove)
    widget.Bind(elistbox.EVT_ELB_EDIT_EVENT,   _listBoxEdit)

    propVal.addListener('widgets_list_py_{}'.format(id(widget)),
                        _listChanged,
                        weak=False)

    return widget


def _List(parent, hasProps, propObj, propVal):

    if propObj.embed:
        return _listEmbedWidget(parent, hasProps, propObj, propVal)
    else:
        return _listDialogWidget(parent, hasProps, propObj, propVal)
